#!/usr/bin/perl
#
# rohttpd: A httpd server.
# Copyright (C) 2006  Roberto Cappellini
#
# This file is part of rohttpd.
#
# rohttpd is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# rohttpd is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with rohttpd; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# =========================================================================

use Socket;
use strict; # e' un pragma per la sicurezza. vedi perldoc -U -m strict
            # mi costringe a dichiarare le variabili com my.

my $proto = getprotobyname('tcp');
my $port = 8080;

# ATTENZIONE! la seguente funzione non e' portabile! pathname Linux.
# copiata da perldoc -U perlipc
use POSIX 'setsid';
sub daemonize
{
  chdir '/'                    or die "err chdir()\n";
  open STDIN, '</dev/null'     or die "err open()\n";
  open STDOUT, '>/dev/null'    or die "err open()\n";
  defined(my $pid = fork())    or die "err fork()\n";
  if ($pid)  { exit(); }
  no strict;
  setsid()                     or die "err setsid()\n";
  use strict;
  open STDERR, '>&STDOUT'      or die "err open()\n";
}

# daemonize(); 

sub timeout
{
  exit 1;
}

# $_[0] $instring
# $_[1] "\r\n"
# Restituisce: ($lineatrovata, $resto) o (undef(), $lineaintera)
# di questa funzione ce n' e' un' altra versione che fa uso
# di index() e che e' da preferire.
# sub rawin
# {
#   if ( $_[0] =~ /$_[1]/)
#     {
#       return ($`, $');
#     }
#   else
#     { return (undef() , $_[0]); }
# }
#
# $_[0] $instring
# $_[1] "\r\n"
# Restituisce: ($lineatrovata, $resto) o (undef(), $lineaintera)
sub rawin
{
  my $pos = 0;
  if (($pos = index($_[0], $_[1], $pos)) >= 0)
    {
      return ((substr $_[0], 0, $pos),
              (substr $_[0], $pos + length($_[1])));
    }
  else
    { return (undef() , $_[0]); }
}

socket(SOCK, PF_INET, SOCK_STREAM, $proto) or
  die "err in syscall socket()\n";
setsockopt(SOCK, SOL_SOCKET, SO_REUSEADDR, pack("l", 1)) or
  die "err in syscall setsockopt()\n";
bind(SOCK, sockaddr_in($port, INADDR_ANY)) or
  die "err in syscall bind()\n";
listen(SOCK, SOMAXCONN) or
  die "err in syscall listen()\n";

$SIG{CHLD} = 'IGNORE'; # Riceve SIGCHLD ed evita popolamento di zombies

my @httpheaders = ();
%main::resources = ();

# ----------- Risorse ---------------------------------------------- #

do './resources.pl';

# ----------- Fine Risorse ----------------------------------------- #

my $keepalive = 4;

print "HTTP Server in esecuzione. Porta $port\n";

while (1)
{
  my $newsocket = accept(REMOTE,SOCK);
  
  if (fork()) { close REMOTE; } # se sono processo padre chiudo
                                # nuovo socket e mi rimetto in accept()
  else
    { # se sono processo figlio chiudo socket in LISTEN e comunico
      # con client tramite nuovo socket.
      close SOCK;
      $SIG{PIPE} = 'IGNORE';
      $SIG{ALRM} = \&timeout;
      
      my ($remoteport, $remoreaddr) = sockaddr_in($newsocket);

      my $incomplete = "";
      my $instring = "";
      
      alarm 30;
      
      INPUT: while (sysread REMOTE, $instring, 8192)
        {
        
          # print '-<< ',$instring,' >>-', "\n";
          syswrite STDOUT, '-<< '.$instring.' >>-'. "\n";
          $instring = $incomplete . $instring;
        
          while (defined((($instring, $incomplete) = rawin $instring, "\r\n")[0]))
	    {
	      if ($instring eq "")
	        {
		  syswrite STDOUT, "--> fine headers\n";
		  
		  alarm 0;
                  my $httpcommand = shift @httpheaders;
                  ($httpcommand, my $resource, my $httpver) =
                    split(/\s/, $httpcommand, 3);
                  
                  ($resource, my $query) = split(/\?/, $resource, 2);
                  
                  if (defined($main::resources{uc($resource)}))
                    {
                      &{$main::resources{uc($resource)}}
                        (\@httpheaders, $httpcommand, $resource, $httpver,
                         \*REMOTE, $query, $incomplete);
                    }
                  else
                    {
                      # Risorsa non trovata
                      resourcenotfound
                        (\@httpheaders, $httpcommand, $resource, $httpver,
                         \*REMOTE, $query, $incomplete);
                    }
                    
                  last INPUT;
                  
		}
              else
	        {
		  if (scalar(@httpheaders) < 512) # accetto MAX 512 headers
		    { push @httpheaders, $instring; }
		  syswrite STDOUT, "--> ".$instring."\n";
		}
              
              $instring = $incomplete; $incomplete = '';
	    }
          
          alarm 30;
        }
        
      exit 0;
    }
}

